#include <idc.idc>

#define I_MOV 122 // mov edi,edi
#define I_PUSH 143 // push/pop pairs
#define I_POP 134
#define I_NOP 129
#define I_NOT 130
#define I_XCHG 214
#define I_BSWAP 218
#define I_JA  53
#define I_JAE 54
#define I_JB  55
#define I_JBE 56
#define I_JCXZ 58
#define I_JECXZ 59
#define I_JG  62
#define I_JGE 63
#define I_JL  64
#define I_JLE 65
#define I_JNB 68
#define I_JNO 76
#define I_JNP 77
#define I_JNS 78
#define I_JNZ 79
#define I_JO  80
#define I_JP  81
#define I_JS  84
#define I_JZ  85
#define I_JMP 86

// get splices count
//auto ea, maxea, splice_ea, max_splice_ea, addr, offset;
//auto count, jmp_addr;
//
//ea = 0x401000;
//maxea = MaxEA();
//splice_ea = 0x7C0000;
//max_splice_ea = 0x800000;
//addr = ea;
//while (addr < maxea) {
//  addr = FindBinary(addr, SEARCH_DOWN, "E9");
//  if (addr == -1)
//    break;
//  offset = Dword(addr + 1);
//  jmp_addr = offset + addr + 5;
//  if (jmp_addr >= splice_ea && jmp_addr < max_splice_ea) {
//    Message("next addr %x\n", addr);
//    count++;
//  }
//  addr = addr + 5;
//}
//Message("total %d jumps found\n", count);

static get_command(addr, cmd_size) {
  auto cmd_code = Qword(addr);
  if (cmd_size == 1) {
      cmd_code = cmd_code & 0xFF;
  } else if (cmd_size == 2) {
      cmd_code = cmd_code & 0xFFFF;
  } else if (cmd_size == 3) {
      cmd_code = cmd_code & 0xFFFFFF;
  } else if (cmd_size == 4) {
      cmd_code = cmd_code & 0xFFFFFFFF;
  } else if (cmd_size == 5) {
      cmd_code = cmd_code & 0xFFFFFFFFFF;
  } else if (cmd_size == 6) {
      cmd_code = cmd_code & 0xFFFFFFFFFFFF;
  } else {
    Warning("Unsupported instrcution size");
    return -1;
  }
  return cmd_code;
}

static MyMakeCode(code_addr) {
  auto cmd_size = MakeCode(code_addr);
  if (!cmd_size) {
    MakeUnknown(code_addr, 0x10, DOUNK_EXPAND);
    cmd_size = MakeCode(code_addr);
  }
  return cmd_size;
}

static CheckUnknownJccTypes(addr) {
  auto cmd;
  auto base;
  auto cmd_size;
  auto ret;

  ret = 0;
  base = SegStart(addr);
  cmd_size = 0;
  auto max_end_addr = SegEnd(addr);
  Message("First addess is %x\n", base);
  while (addr < max_end_addr) {
    addr = addr + cmd_size;
    cmd_size = MyMakeCode(addr);
    if (Dword(addr) == 0) {
      Message("Last addess is %x\n", addr);
      break;
    }
    cmd = DecodeInstruction(addr);
    if (!cmd) {
      Message("Failed to decode %x!\n", addr);
      ret = -1;
      break;
    }
    if (cmd.itype == 86 && !ret) {
      // get base of spliced code
      if (cmd.Op0.addr < base || cmd.Op0.addr > max_end_addr) {
        ret = SegStart(cmd.Op0.addr);
      }
    }
    if (cmd.itype >= 53 && cmd.itype < 86) {
      if (cmd.itype != 86) {
        auto next_addr = addr + cmd_size;
        if (cmd.Op0.addr == next_addr) {
          continue; // jcc type $+2, $+5 (without jump)
        }
        auto next_cmd_size = MyMakeCode(next_addr);
        // if jcc not like $+2, $+5 (without jump) check
        // if the next cmd is jcc with the same size
        if (next_cmd_size != cmd_size) {
          Message("Found unknown sequence %x %x %x!\n", addr, next_addr, next_cmd_size);
          ret = -1;
          break;
        }
        auto next_cmd = DecodeInstruction(next_addr);
        if (next_cmd.itype >= 53 && next_cmd.itype < 86) {
          // check for pair jmp
          if (next_cmd.itype != cmd.itype) {
            Message("Found unknown sequence %x, addr type %d!\n", addr, next_cmd.itype);
            ret = -1;
            break;
          }
          if (cmd.Op0.addr != next_addr + next_cmd_size) {
            Message("Found unknown sequence %x, jump to %x, %x !\n", addr, cmd.Op0.addr, next_addr);
            ret = -1;
            break;
          }
          // process double jcc as block
          addr = addr + cmd_size;
        } else {
          Message("Found unknown sequence %x, jump to %x!\n", addr, cmd.Op0.addr);
          ret = -1;
          break;
        }
      }
    }
  }
  return ret;
}

static get_next_splice(addr, splice_ea, max_splice_ea) {
  auto jmp_addr;
  auto cmd;
  auto max_code_addr;

  max_code_addr = SegEnd(addr);
  while (addr < max_code_addr) {
    addr = FindBinary(addr, SEARCH_DOWN, "E9");
    if (addr == -1)
      break;
    cmd = DecodeInstruction(addr);
    if (!cmd) {
      Message("Failed to decode %x!\n", addr);
      addr++;
      continue;
    }
    jmp_addr = cmd.Op0.addr;
    if (jmp_addr >= splice_ea && jmp_addr < max_splice_ea) {
      //Message("next addr %x\n", addr);
      return addr;
    }
    addr = addr + 5;
  }
  return 0;
}

/** check instruction type
  * @return
  *  o if instruction is not known for deobfuscator
  *  1 on single 'No Operation' instruction
  *  2 on paired instruction (can be obfuscation, can be not)
  */
static check_instruction_type(addr, cmd, cmd_size) {
  auto result = 0;

  // check BSWAP pair
  if (cmd.itype == I_BSWAP) {
    result = 2;
  }
  // check PUSH/POP pair
  else if (cmd.itype == I_PUSH) {
    result = 2;
  }
  // check NOT pair
  else if (cmd.itype == I_NOT) {
    result = 2;
  }
  // check XCHG pair
  else if (cmd.itype == I_XCHG) {
    if (cmd.Op0.reg == cmd.Op1.reg) {
      result = 1;
    } else
      result = 2;
  }
  // check NOP
  else if (cmd.itype == I_NOP) {
    result = 1;
  }
  // check MOV RX,RX
  else if (cmd.itype == I_MOV) {
    if (cmd.Op0.type == o_reg && cmd.Op1.type == o_reg) {
      if (cmd.Op0.reg == cmd.Op1.reg) {
        result = 1;
      }
    }
  }
  // Jcc $+next
  if (cmd.itype >= 53 && cmd.itype <= 86) {
    if (cmd.Op0.addr == addr + cmd_size) {
      result = 1;
    }
  }
  return result;
}

static check_pair_block(analyze_block_addr) {
  // search for the next occurance
  auto check_pair = 0;
  auto pair_itype, pair_cmd_code;
  auto addr = analyze_block_addr;
  auto max_addr = SegEnd(addr);
  auto pair_reg, auxpref;
  auto next_addr;
  auto found = 0;
  auto cmd, cmd_size, cmd_type;

  cmd_size = MakeCode(addr);
  if (!cmd_size) {
    Message("Failed MakeCode %x!\n", cmd_size);
    return -1;
  }
  cmd = DecodeInstruction(addr);
  if (!cmd) {
    Message("Failed to decode!\n");
    return -1;
  }
  if (cmd.Op0.type != o_reg) {
    //Message("Invalid type of operand!\n");
    return -1;
  }
  // check type of instruction
  cmd_type = check_instruction_type(addr, cmd, cmd_size);
  if (cmd_type != 2) {
    Message("cmd_type %d\n", cmd_type);
    return -1;
  }
  pair_reg = cmd.Op0.reg;
  auxpref = cmd.auxpref;
  // get paired instruction type
  if (cmd.itype == I_PUSH) {
    pair_itype = I_POP;
  } else {
    pair_itype = cmd.itype;
  }
  addr = addr + cmd_size;
  while (addr < max_addr) {
    cmd_size = MakeCode(addr);
    if (!cmd_size) {
      Message("Failed MakeCode %x!\n", cmd_size);
      return -1;
    }
    cmd = DecodeInstruction(addr);
    if (!cmd) {
      Message("Failed to decode!\n");
      return -1;
    }
    // check type of instruction
    cmd_type = check_instruction_type(addr, cmd, cmd_size);
    if (cmd_type == 1) {
      // NOPs, non standart NOPs, MOV RX,RX, fake jumps
      addr = addr + cmd_size;
      continue;
    }
    // check pair
    if (cmd.itype == pair_itype) {
      // check if register is same
      if (pair_reg == cmd.Op0.reg && cmd.auxpref == auxpref) {
        next_addr = addr + cmd_size;
        found = 1;
        break;
      }
      if (cmd.itype == I_POP) {
        // if we here a POP instruction is not obfuscation
        return -1;
      }
      // other cases will be processed at the end of loop
    }
    if (cmd_type == 2) {
      // check pairs
      auto block_size = check_pair_block(addr);
      if (block_size > 0) {
        addr = addr + block_size;
        continue;
      } else {
        // block analyze failed
        return -1;
      }
    }
    if (cmd.itype == I_JMP) {
      return -1;
    }
    next_addr = addr + cmd_size;
    // check pair of Jcc
    if (cmd.itype >= 53 && cmd.itype < 86) {
      auto new_cmd_size = MakeCode(next_addr);
      if (!new_cmd_size) {
        Message("Failed MakeCode %x!\n", next_addr);
        return -1;
      }
      auto new_cmd = DecodeInstruction(next_addr);
      if (!new_cmd) {
        Message("Failed to decode %x!\n", next_addr);
        return -1;
      }
      // pair of Jcc
      if (new_cmd.itype == cmd.itype) {
        addr = addr + cmd_size;
        addr = addr + new_cmd_size;
        continue;
      } else {
        // unknown Jcc (not obfuscation)
        return -1;
      }
    }
    // unknown non-paired instruction, block analyze failed
    return -1;
  }
  if (found) {
    //Message("code from %x to %x is obfuscation block!\n", analyze_block_addr, addr);
    return next_addr - analyze_block_addr;
  }
  return -1;
}

static calc_splice_max_size(splice_addr, needColors) {
  auto addr, max_addr, next_addr;
  auto jmp_addr;
  auto size;
  auto cmd, cmd_size;
  auto is_pair_instr;
  auto pair_itype;
  auto pair_cmd_code;
  auto cmd_type;

  cmd_size = MyMakeCode(splice_addr);
  if (cmd_size != 5) {
    Message("Invalid adress for splice %x", splice_addr);
    return -1;
  }
  addr = 0;
  size = 5;
  max_addr = SegEnd(splice_addr);
  // check for obfuscation cmds
  while (addr < max_addr) {
    addr = splice_addr + size;
    cmd_size = MyMakeCode(addr);
    if (!cmd_size) {
      Message("Failed MakeCode %x!\n", addr);
      return -1;
    }
    cmd = DecodeInstruction(addr);
    if (!cmd) {
      Message("Failed to decode %x!\n", addr);
      return -1;
    }
    // check type of instruction
    cmd_type = check_instruction_type(addr, cmd, cmd_size);
    if (cmd_type == 1) {
      // NOPs, non standart NOPs, MOV RX,RX, fake jumps
      size = size + cmd_size;
      continue;
    }
    // check pairs
    if (cmd_type == 2) {
      //Message("check block %x\n", addr);
      auto block_size = check_pair_block(addr);
      //Message("block size %x\n", block_size);
      if (block_size > 0) {
        size = size + block_size;
        continue;
      }
    }
    // JMP can be used
    if (cmd.itype == I_JMP) {
      return size;
    }
    next_addr = addr + cmd_size;
    // check Jcc
    if (cmd.itype >= 53 && cmd.itype < 86) {
      auto new_cmd_size = MakeCode(next_addr);
      if (!new_cmd_size) {
        Message("Failed MakeCode %x!\n", next_addr);
        return -1;
      }
      auto new_cmd = DecodeInstruction(next_addr);
      if (!new_cmd) {
        Message("Failed to decode %x!\n", next_addr);
        return -1;
      }
      // pair of Jcc
      if (new_cmd.itype == cmd.itype) {
        size = size + cmd_size;
        size = size + new_cmd_size;
        continue;
      } else {
        // unknown Jcc (not obfuscation)
        return -2;
      }
    }
    // next instructions unknown
    break;
  }
  if (needColors) {
    auto i;
    addr = splice_addr + size;
    for (i = splice_addr; i < addr; i = i + MakeCode(i)) {
      SetColor(i, CIC_ITEM, 0x55FF80);
    }
  }
  return size;
}

static calc_splice_possible_size(splice_addr) {
  auto addr, min_addr, max_addr, next_addr;
  auto jmp_addr;
  auto size;
  auto cmd, cmd_size;
  auto is_pair_instr;
  auto pair_itype;
  auto pair_cmd_code;
  auto splice_start_addr;
  auto cmd_type;

  cmd_size = MyMakeCode(splice_addr);
  if (cmd_size != 5) {
    Message("Invalid adress for splice %x", splice_addr);
    return -1;
  }
  cmd = DecodeInstruction(splice_addr);
  if (!cmd) {
    Message("Failed to decode %x!\n", splice_addr);
    return -1;
  }
  if (cmd.itype != I_JMP) {
    Message("Invalid adress for splice %x", splice_addr);
  }
  addr = cmd.Op0.addr;
  size = 0;
  cmd_size = 0;
  min_addr = SegStart(addr);
  max_addr = SegEnd(addr);
  // check for obfuscation cmds
  while (addr < max_addr) {
    cmd_size = MyMakeCode(addr);
    if (!cmd_size) {
      Message("Failed MakeCode %x!\n", addr);
      return -1;
    }
    cmd = DecodeInstruction(addr);
    if (!cmd) {
      Message("Failed to decode %x!\n", addr);
      return -1;
    }
    // check type of instruction
    cmd_type = check_instruction_type(addr, cmd, cmd_size);
    if (cmd_type == 1) {
      // NOPs, non standart NOPs, MOV RX,RX, fake jumps
      SetColor(addr, CIC_ITEM, 0x55FF80);
      addr = addr + cmd_size;
      continue;
    }
    // check pairs
    if (cmd_type == 2) {
      //Message("check block %x\n", addr);
      auto block_size = check_pair_block(addr);
      //Message("block size %x\n", block_size);
      if (block_size > 0) {
        auto i, block_addr = addr;
        addr = addr + block_size;
        for (i = block_addr; i < addr; i = i + MakeCode(i)) {
          SetColor(i, CIC_ITEM, 0x55FF80);
        }
        continue;
      }
    }
    // JMP can be used
    if (cmd.itype == I_JMP) {
      if (cmd.Op0.addr < min_addr || cmd.Op0.addr > max_addr) {
        SetColor(addr, CIC_ITEM, 0x55FF80);
        return size;
      } else {
        Message("Unknown JMP %x\n", addr);
        return -1;
      }
    }
    next_addr = addr + cmd_size;
    // check Jcc
    if (cmd.itype >= 53 && cmd.itype < 86) {
      auto new_cmd_size = MakeCode(next_addr);
      if (!new_cmd_size) {
        Message("Failed MakeCode %x!\n", next_addr);
        return -1;
      }
      auto new_cmd = DecodeInstruction(next_addr);
      if (!new_cmd) {
        Message("Failed to decode %x!\n", next_addr);
        return -1;
      }
      // pair of Jcc
      if (new_cmd.itype == cmd.itype) {
        SetColor(addr, CIC_ITEM, 0x55FF80);
        addr = addr + cmd_size;
        SetColor(addr, CIC_ITEM, 0x55FF80);
        addr = addr + new_cmd_size;
        continue;
      } else {
        // unknown Jcc (not obfuscation)
        return -2;
      }
    }
    // next instructions unknown
    size = size + cmd_size;
    addr = addr + cmd_size;
  }
  return size;
}

static patch_splice(splice_addr, splice_size) {
  auto addr, min_addr, max_addr, next_addr;
  auto jmp_addr;
  auto cmd, cmd_size;
  auto is_pair_instr;
  auto pair_itype;
  auto pair_cmd_code;
  auto splice_start_addr;
  auto cmd_type;
  auto write_ptr = splice_addr;
  auto i;

  cmd_size = MyMakeCode(splice_addr);
  if (cmd_size != 5) {
    Message("Invalid adress for splice %x", splice_addr);
    return -1;
  }
  cmd = DecodeInstruction(splice_addr);
  if (!cmd) {
    Message("Failed to decode %x!\n", splice_addr);
    return -1;
  }
  if (cmd.itype != I_JMP) {
    Message("Invalid adress for splice %x", splice_addr);
  }
  addr = cmd.Op0.addr;
  cmd_size = 0;
  min_addr = SegStart(addr);
  max_addr = SegEnd(addr);
  // check for obfuscation cmds
  while (addr < max_addr) {
    cmd_size = MyMakeCode(addr);
    if (!cmd_size) {
      Message("Failed MakeCode %x!\n", addr);
      return -1;
    }
    cmd = DecodeInstruction(addr);
    if (!cmd) {
      Message("Failed to decode %x!\n", addr);
      return -1;
    }
    // check type of instruction
    cmd_type = check_instruction_type(addr, cmd, cmd_size);
    if (cmd_type == 1) {
      // NOPs, non standart NOPs, MOV RX,RX, fake jumps
      addr = addr + cmd_size;
      continue;
    }
    // check pairs
    if (cmd_type == 2) {
      auto block_size = check_pair_block(addr);
      if (block_size > 0) {
        addr = addr + block_size;
        continue;
      }
    }
    // JMP can be used
    if (cmd.itype == I_JMP) {
      if (cmd.Op0.addr < min_addr || cmd.Op0.addr > max_addr) {
        break;
      } else {
        Message("Unknown JMP %x\n", addr);
        return -1;
      }
    }
    next_addr = addr + cmd_size;
    // check Jcc
    if (cmd.itype >= 53 && cmd.itype < 86) {
      auto new_cmd_size = MakeCode(next_addr);
      if (!new_cmd_size) {
        Message("Failed MakeCode %x!\n", next_addr);
        return -1;
      }
      auto new_cmd = DecodeInstruction(next_addr);
      if (!new_cmd) {
        Message("Failed to decode %x!\n", next_addr);
        return -1;
      }
      // pair of Jcc
      if (new_cmd.itype == cmd.itype) {
        addr = addr + cmd_size;
        addr = addr + new_cmd_size;
        continue;
      } else {
        // unknown Jcc (not obfuscation)
        return -2;
      }
    }
    // repair instruction
    for (i = 0; i < cmd_size; ++i) {
      PatchByte(write_ptr + i, Byte(addr + i));
    }
    addr = addr + cmd_size;
    write_ptr = write_ptr + cmd_size;
  }
  MyMakeCode(splice_addr);
  return 1;
}

static main() {
  auto code_addr, splice_addr = 0x3B0000;
  auto count = 0, patched = 0;
  auto free_size, splice_size;
  
  splice_addr = AskAddr(splice_addr, "Enter address of splices");
  Message("===============Start===============\n");
  // search for code segment and check Jcc/JMP types
  code_addr = CheckUnknownJccTypes(splice_addr);
  // prepare splice address constants
  auto splice_ea = SegStart(splice_addr);
  auto max_splice_ea = SegEnd(splice_ea);
  Message("Code offset %x, splice first ea %x, end ea %x\n", code_addr, splice_ea, max_splice_ea);
  while (1) {
    code_addr = get_next_splice(code_addr, splice_ea, max_splice_ea);
    if (!code_addr)
      break;
    count++;
    free_size = calc_splice_max_size(code_addr, 0);
    splice_size = calc_splice_possible_size(code_addr);
    //Message("Free space size %x, block size %x\n", free_size, splice_size);
    if (free_size != splice_size || free_size < 5) {
      Message("Free space size %x, block size %x\n", free_size, splice_size);
      Message("Sizes differs! Check address %x manually!\n", code_addr);
    } else {
      if (patch_splice(code_addr, free_size) == 1)
        patched++;
    }
    code_addr = code_addr + 5;
  }
  Message("total %d jumps found and %d patched!\n", count, patched);
}
